// ######### Software Disclaimer #############
//This software is provided free of charge to the academic research community for non-commercial research and educational purposes only. For information on the use for a commercial purpose or by a commercial or for-profit entity, please contact Prof. Xiangong Jasmine Zhou (WWW: https://zhoulab.dgsom.ucla.edu/, and Email: xjzhou@mednet.ucla.edu).
//
//CancerDetector is copyrighted by the University of California, Los Angeles (UCLA). It can be freely used for educational and research purposes by non-profit institutions and U.S. government agencies only. In accordance with the terms herein, UCLA grants to LICENSEE and LICENSEE accepts from UCLA, a non-exclusive, non-transferable license to use the Software for LICENSEE's evaluation and internal research purposes only. LICENSEE recognizes and understands that it shall have no rights whatsoever to distribute the Software unless such rights are granted to LICENSEE by UCLA in a separate agreement.
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
//

#include <cstdlib>
#include <vector>
#include <map>
#include <string>
#include <climits>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cassert>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_exp.h>

using namespace std;

typedef unsigned int GENOME_POSITION;
// Bins2Values: a map for bin_index -> two vectors of values. bin_index is always 1-base
typedef map<int, pair<vector<double>,vector<double> > > Bins2PairedValues;

vector<string> split(string str, char delimiter)
{
	vector<string> items;
	istringstream ss( str );
	string token;
	while(getline(ss, token, delimiter)) {
		items.push_back(token);
	}
	return items;
}

// Read marker file
// Each line is a marker bin (or feature). All columns are delimited by TAB. There is one header line.
// Column 1: marker_index, 1-based index. Only marker bins are included, those complementary bins do not appear in this file.
// Column 2: chr
// Column 3: start coordinate of bin (1-base)
// Column 4: end coordinate of bin (1-base). The range of the bin is [start, end)
// Column 5+: paired values for this marker (each column is a class). The first class is tumor and the second is normal class. Each class has a pair of values (alpha and beta shape parameters of a beta distribution), which are delimited by ":"

//For example:
//marker_index	chr	start	end	tumor	normal_plasma
//1	chr1	1	402	3.89692838577676:6.41446579743991	37.6159266954451:12.0867424085961
//2	chr1	101359	101646	2.87841390045233:2.93349657854305	36.4032438100824:3.1110956610252
//3	chr1	102391	102860	2.84669435249341:2.76935335266805	31.3421864311187:3.7799000264534
//4	chr1	113512	114254	3.17828827147782:2.688981888202	59.6678114550981:7.71569189859518
//5	chr1	473277	473599	4.15708661124289:2.71858405253333	3.38883865010887:32.5214534266035
//...
//
// input file is TAB-delimited plain text. The file has a header line
// value_column_start_index: the column index of the first value (index is 1-base). In this format, it should be 5.
// bins2pairedvalues: a map for bin_index -> a pair of two vectors of values. bin_index is always 1-base. In detail, it is "map<int, pair<vector<double>,vector<double> > >"
void read_paired_values_file_of_bins(string value_file_of_bins, unsigned int value_column_start_index,
	Bins2PairedValues & bins2pairedvalues, vector<string> & value_names)
{
	unsigned long i_line=0;
	// input is a plain text file
	ifstream fin;
	fin.open(value_file_of_bins.c_str());
	if (fin.fail()){
		cerr << "Error: Unable to open " << value_file_of_bins << " in read_paired_values_file_of_bins()" << endl;
		exit(EXIT_FAILURE);
	}
	string line;
	while (!fin.eof()) {
		getline(fin, line);
		if (i_line==0) {
			// skip the header line of value_file_of_bins
			vector<string> items = split(line, '\t');
		
			for (int i=value_column_start_index-1; i<items.size(); i++) {
				value_names.push_back( items[i] );
			}
			i_line++;
			continue;
		}
		if (line.empty()) {
			// this is the last line of the file
			break;
		}
		vector<string> items = split(line, '\t');
		if (value_column_start_index>items.size()) {
			cerr << "Error(read_paired_values_file_of_bins): the value column starts from Column " << value_column_start_index << " that is > total number of columns (" << items.size() << ") in Line " << i_line+1 << "!" << endl;
			exit(EXIT_FAILURE);
		}
		int marker_index = (int)atoi(items[0].c_str());
		vector<double> values1, values2;
		for (int i=value_column_start_index-1; i<items.size(); i++) {
			vector<string> subitems = split(items[i], ':');
			if (subitems.size()!=2) {
				cerr << "Error(read_paired_values_file_of_bins): the value column " << (i+1) << " has no paired values (only " << subitems.size() << " values) in Line " << i_line+1 << "!" << endl;
				exit(EXIT_FAILURE);
			}
			values1.push_back( (double)atof(subitems[0].c_str()) );
			values2.push_back( (double)atof(subitems[1].c_str()) );
		}
		bins2pairedvalues[marker_index] = make_pair(values1, values2);
		i_line++;
	}
	fin.close();
}

//
//input file of indexing reads by bins (or features). (generated by program "reads_binning")
//Each line is a read. All columns are delimited by TAB. There is one header line.
//Column 1: marker1_index, 1-based index. "0" indicates the bin "marker1_index" is not a marker.
//Column 2: marker2_index, 1-based index. "-" indicates the read completely falls into "marker1"; "0" indicates the bin "marker2_index" is not a marker.
//Column 3: number of CpG sites
//Column 4: a binary vector of methylation status for all CpG sites in this read (no delimit). This vector should have the same size as list size in Column 6. For example, 00111100
//
//For example:
//marker1_index	marker2_index	num_CpG_sites	sites_methylation_status
//1	-	3	100
//1	-	6	111111
//1	-	6	010111
//1	-	5	11111
//11	-	6	111111
//11	-	3	000
//11	-	7	0111111
//11	-	8	11111111
//12	-	4	1101
//12	-	8	11111111
//12	-	8	11111111
//12	0	7	1111111
//12	0	6	011111
//0	13	11	111
//13	-	10	1111111111
//13	-	3	000
//13	-	3	101
//...
//
// methy_status_column_index is 1-base
void process_one_line_of_reads_binning_file(string & line, int methy_status_column_index, int & marker_index, vector<int> & methy_status)
{
	vector<string> items = split(line, '\t');
	marker_index = (unsigned int)atoi(items[0].c_str());
	if (marker_index==0) {
		// indicates this read covers two bins: a complementary bin (marker_index=0) and a marker bin (marker_index is non-zero).
		// so we should take the second item as the marker_index
		// for example:
		// 0	2	3	101
		marker_index = (unsigned int)atoi(items[1].c_str());
	}
	string & str_methy_status = items[methy_status_column_index-1];
	for (int i=0; i<str_methy_status.size(); i++)
		methy_status.push_back(str_methy_status.at(i)-'0'); // convert each char '0' or '1' to 0 or 1, respectively.
}

// Function: calculate the likelihood of each read that belongs to a class
//    the formula is B(x+alpha, 1-x+beta)/B(alpha,beta)
//    where B() is beta function, x is methylation status of a CpG site (0 or 1), alpha and beta are a pair of values for describing a Beta-distribution provided by the input parameter "bins2pairedvalues".
//
// background_B: a precomputed B(alpha,beta)
//
double calc_one_read_prob_by_a_paired_value(double alpha, double beta, double background_B, vector<int> & methy_status)
{
	if (!methy_status.empty()) {
		double p=1.0;
		for (int i=0; i<methy_status.size(); i++) {
			// for each CpG site
			p *= gsl_sf_exp( gsl_sf_lngamma(methy_status[i]+alpha) + gsl_sf_lngamma(1-methy_status[i]+beta) - gsl_sf_lngamma(alpha+beta+1) - background_B );
		}
		return p;
	} else {
		return -1;
	}
}
//
// bins2pairedvalues: a map for bin_index -> a pair of two vectors of value. bin_index is always 1-base.
//
// Function: calculate the likelihood of each read that belongs to a class
//    the formula is B(x+alpha, 1-x+beta)/B(alpha,beta)
//    where B() is beta function, x is methylation status of a CpG site (0 or 1), alpha and beta are a pair of values for describing a Beta-distribution provided by the input parameter "bins2pairedvalues".
//
// Output: class-specific likelihood of each read
//
//Each line is a read. All columns are delimited by TAB. There is one header line.
//Column 1: marker index (1-base)
//Columns 2+: each column is a probability value of a class. For example, given 2 classes, there will be 2 additional columns.
void calc_read_probability_by_bins2pairedvalues(string reads_binning_file,
	Bins2PairedValues & bins2pairedvalues, vector<string> & value_names)
{
	int column_index_of_methy_status = 4; // 1-base index
	// precompute the B(alpha,beta) for each bin
	map<int,vector<double> > B;
	Bins2PairedValues::iterator it;
	for (it=bins2pairedvalues.begin(); it!=bins2pairedvalues.end(); ++it) {
		int bin_index = it->first;
		vector<double> & alphas = (it->second).first;
		vector<double> & betas = (it->second).second;
		vector<double> values;
		int n = alphas.size();
		for (int t=0; t<n; t++) {
			values.push_back( gsl_sf_lngamma(alphas[t]) + gsl_sf_lngamma(betas[t]) - gsl_sf_lngamma(alphas[t]+betas[t]) );
		}
		B[bin_index] = values;
	}
	cout.precision(15);
	// print the header line
	cout << "marker_index";
	for (int i=0; i<value_names.size(); i++)
		cout << "\t" << value_names[i];
	cout << "\n";
	// process each read (i.e., line) of the input file, then print out the calculated probabilities for each read.
	unsigned long i_line=0; // line count of the input file
	istream * in=&cin; // default is stdin
	ifstream fin;
	if ( !(reads_binning_file.compare("stdin")==0)) {
		fin.open(reads_binning_file.c_str());
		if (fin.fail()){
			cerr << "Error: Unable to open " << reads_binning_file << " in calc_read_probability_by_bins2values()" << endl;
			exit(EXIT_FAILURE);
		}
		in = &fin;
	}
	string line;
	while (!(*in).eof()) {
		getline((*in), line);
		if (i_line==0) {
			// skip the header line of wig file
			i_line++;
			continue;
		}
		if (line.empty()) {
			// this is the last line of the file
			break;
		}
		int marker_index;
		vector<int> methy_status;
		process_one_line_of_reads_binning_file(line, column_index_of_methy_status, marker_index, methy_status);
		if (!(bins2pairedvalues.find(marker_index) == bins2pairedvalues.end())) {
			cout << marker_index; 
			vector<double> & alphas = bins2pairedvalues[marker_index].first;
			vector<double> & betas = bins2pairedvalues[marker_index].second;
			for (int t=0; t<alphas.size(); t++) {
				double p = calc_one_read_prob_by_a_paired_value(alphas[t], betas[t], B[marker_index][t], methy_status);
				cout << "\t" << p; 
			}
			cout << "\n";
		}
		i_line++;
	}
	if ( !(reads_binning_file.compare("stdin")==0) ) {
		fin.close();
	}
}

int main(int argc, char* argv[]) {
	if (argc != 3) {
		cerr << "USAGE: CalcReadLikelihood [input reads file] [input markers file]" << endl;
		exit(EXIT_FAILURE);
	}
	string reads_file = argv[1];
	string markers_file = argv[2];

	unsigned int value_column_start_index=5; // column index is 1-base. This is the first value's column index of the file "value_file_of_bins"
	Bins2PairedValues bins2pairedvalues;
	vector<string> class_names;
	read_paired_values_file_of_bins(markers_file, value_column_start_index, bins2pairedvalues, class_names);
	calc_read_probability_by_bins2pairedvalues(reads_file, bins2pairedvalues, class_names);
}

